@@ -2,6 +2,8 @@
 #include "internalSharedTypes.hpp"
 #include "internalDefines.hpp"
 #include <array>
+#include <sstream>
+#include <cstdio>
 #include <filesystem>
 #include <zip.h>
 #include <cstring>
@@ -15,11 +17,25 @@
 
 using namespace Hyprcursor;
 
-// directories for lookup
-constexpr const std::array<const char*, 1> systemThemeDirs = {"/usr/share/icons"};
+static std::vector<std::string> getSystemThemeDirs() {
+    const auto               envXdgData = std::getenv("XDG_DATA_DIRS");
+    std::vector<std::string> result;
+    if (envXdgData) {
+        std::stringstream envXdgStream(envXdgData);
+        std::string       tmpStr;
+        while (getline(envXdgStream, tmpStr, ':'))
+            result.push_back((tmpStr + "/icons"));
+    } else
+        result = {"/usr/share/icons"};
+
+    return result;
+}
+
+const std::vector<std::string>             systemThemeDirs = getSystemThemeDirs();
 constexpr const std::array<const char*, 2> userThemeDirs   = {"/.local/share/icons", "/.icons"};
 
 //
+
 static std::string themeNameFromEnv(PHYPRCURSORLOGFUNC logfn) {
     const auto ENV = getenv("HYPRCURSOR_THEME");
     if (!ENV) {
@@ -291,13 +307,15 @@ SCursorImageData** CHyprcursorManager::getShapesC(int& outSize, const char* shap
         if (REQUESTEDSHAPE != shape->directory && std::find(shape->overrides.begin(), shape->overrides.end(), REQUESTEDSHAPE) == shape->overrides.end())
             continue;
 
+        const int PIXELSIDE = std::round(info.size / shape->nominalSize);
+
         hotX = shape->hotspotX;
         hotY = shape->hotspotY;
 
         // matched :)
         bool foundAny = false;
         for (auto& image : impl->loadedShapes[shape.get()].images) {
-            if (image->side != info.size)
+            if (image->side != PIXELSIDE)
                 continue;
 
             // found size
@@ -317,7 +335,7 @@ SCursorImageData** CHyprcursorManager::getShapesC(int& outSize, const char* shap
         // find nearest
         int leader = 13371337;
         for (auto& image : impl->loadedShapes[shape.get()].images) {
-            if (std::abs((int)(image->side - info.size)) > std::abs((int)(leader - info.size)))
+            if (std::abs((int)(image->side - PIXELSIDE)) > std::abs((int)(leader - PIXELSIDE)))
                 continue;
 
             leader = image->side;
@@ -373,14 +391,25 @@ SCursorRawShapeDataC* CHyprcursorManager::getRawShapeDataC(const char* shape_) {
 
     SCursorRawShapeDataC*            data = new SCursorRawShapeDataC;
     std::vector<SLoadedCursorImage*> resultingImages;
+    data->overridenBy = nullptr;
+    data->images      = nullptr;
+    data->len         = 0;
+    data->hotspotX    = 0.f;
+    data->hotspotY    = 0.F;
+    data->nominalSize = 1.F;
+    data->resizeAlgo  = eHyprcursorResizeAlgo::HC_RESIZE_NONE;
+    data->type        = eHyprcursorDataType::HC_DATA_PNG;
 
     for (auto& shape : impl->theme.shapes) {
         // if it's overridden just return the override
-        if (const auto IT = std::find(shape->overrides.begin(), shape->overrides.end(), SHAPE); IT != shape->overrides.end()) {
-            data->overridenBy = strdup(IT->c_str());
+        if (const auto IT = std::find_if(shape->overrides.begin(), shape->overrides.end(), [&](const auto& e) { return e == SHAPE && SHAPE != shape->directory; });
+            IT != shape->overrides.end()) {
+            data->overridenBy = strdup(shape->directory.c_str());
             return data;
         }
+    }
 
+    for (auto& shape : impl->theme.shapes) {
         if (shape->directory != SHAPE)
             continue;
 
@@ -392,9 +421,10 @@ SCursorRawShapeDataC* CHyprcursorManager::getRawShapeDataC(const char* shape_) {
             resultingImages.push_back(i.get());
         }
 
-        data->hotspotX = shape->hotspotX;
-        data->hotspotY = shape->hotspotY;
-        data->type     = shape->shapeType == SHAPE_PNG ? HC_DATA_PNG : HC_DATA_SVG;
+        data->hotspotX    = shape->hotspotX;
+        data->hotspotY    = shape->hotspotY;
+        data->nominalSize = shape->nominalSize;
+        data->type        = shape->shapeType == SHAPE_PNG ? HC_DATA_PNG : HC_DATA_SVG;
         break;
     }
 
@@ -424,8 +454,10 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
         bool sizeFound = false;
 
         if (shape->shapeType == SHAPE_PNG) {
+            const int IDEALSIDE = std::round(info.size / shape->nominalSize);
+
             for (auto& image : impl->loadedShapes[shape.get()].images) {
-                if (image->side != info.size)
+                if (image->side != IDEALSIDE)
                     continue;
 
                 sizeFound = true;
@@ -439,7 +471,7 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
             SLoadedCursorImage* leader    = nullptr;
             int                 leaderVal = 1000000;
             for (auto& image : impl->loadedShapes[shape.get()].images) {
-                if (image->side < info.size)
+                if (image->side < IDEALSIDE)
                     continue;
 
                 if (image->side > leaderVal)
@@ -451,7 +483,7 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
 
             if (!leader) {
                 for (auto& image : impl->loadedShapes[shape.get()].images) {
-                    if (std::abs((int)(image->side - info.size)) > leaderVal)
+                    if (std::abs((int)(image->side - IDEALSIDE)) > leaderVal)
                         continue;
 
                     leaderVal = image->side;
@@ -468,12 +500,16 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
 
             Debug::log(HC_LOG_TRACE, logFn, "loadThemeStyle: png shape {} has {} frames", shape->directory, FRAMES.size());
 
+            const int PIXELSIDE = std::round(leader->side / shape->nominalSize);
+
+            Debug::log(HC_LOG_TRACE, logFn, "loadThemeStyle: png shape has nominal {:.2f}, pixel size will be {}x", shape->nominalSize, PIXELSIDE);
+
             for (auto& f : FRAMES) {
                 auto& newImage           = impl->loadedShapes[shape.get()].images.emplace_back(std::make_unique<SLoadedCursorImage>());
                 newImage->artificial     = true;
-                newImage->side           = info.size;
-                newImage->artificialData = new char[info.size * info.size * 4];
-                newImage->cairoSurface   = cairo_image_surface_create_for_data((unsigned char*)newImage->artificialData, CAIRO_FORMAT_ARGB32, info.size, info.size, info.size * 4);
+                newImage->side           = PIXELSIDE;
+                newImage->artificialData = new char[PIXELSIDE * PIXELSIDE * 4];
+                newImage->cairoSurface   = cairo_image_surface_create_for_data((unsigned char*)newImage->artificialData, CAIRO_FORMAT_ARGB32, PIXELSIDE, PIXELSIDE, PIXELSIDE * 4);
                 newImage->delay          = f->delay;
 
                 const auto PCAIRO = cairo_create(newImage->cairoSurface);
@@ -487,12 +523,12 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
 
                 const auto PTN = cairo_pattern_create_for_surface(f->cairoSurface);
                 cairo_pattern_set_extend(PTN, CAIRO_EXTEND_NONE);
-                const float scale = info.size / (float)f->side;
+                const float scale = PIXELSIDE / (float)f->side;
                 cairo_scale(PCAIRO, scale, scale);
                 cairo_pattern_set_filter(PTN, shape->resizeAlgo == HC_RESIZE_BILINEAR ? CAIRO_FILTER_GOOD : CAIRO_FILTER_NEAREST);
                 cairo_set_source(PCAIRO, PTN);
 
-                cairo_rectangle(PCAIRO, 0, 0, info.size, info.size);
+                cairo_rectangle(PCAIRO, 0, 0, PIXELSIDE, PIXELSIDE);
 
                 cairo_fill(PCAIRO);
                 cairo_surface_flush(newImage->cairoSurface);
@@ -505,12 +541,16 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
 
             Debug::log(HC_LOG_TRACE, logFn, "loadThemeStyle: svg shape {} has {} frames", shape->directory, FRAMES.size());
 
+            const int PIXELSIDE = std::round(info.size / shape->nominalSize);
+
+            Debug::log(HC_LOG_TRACE, logFn, "loadThemeStyle: svg shape has nominal {:.2f}, pixel size will be {}x", shape->nominalSize, PIXELSIDE);
+
             for (auto& f : FRAMES) {
                 auto& newImage           = impl->loadedShapes[shape.get()].images.emplace_back(std::make_unique<SLoadedCursorImage>());
                 newImage->artificial     = true;
-                newImage->side           = info.size;
-                newImage->artificialData = new char[info.size * info.size * 4];
-                newImage->cairoSurface   = cairo_image_surface_create_for_data((unsigned char*)newImage->artificialData, CAIRO_FORMAT_ARGB32, info.size, info.size, info.size * 4);
+                newImage->side           = PIXELSIDE;
+                newImage->artificialData = new char[PIXELSIDE * PIXELSIDE * 4];
+                newImage->cairoSurface   = cairo_image_surface_create_for_data((unsigned char*)newImage->artificialData, CAIRO_FORMAT_ARGB32, PIXELSIDE, PIXELSIDE, PIXELSIDE * 4);
                 newImage->delay          = f->delay;
 
                 const auto PCAIRO = cairo_create(newImage->cairoSurface);
@@ -528,7 +568,7 @@ bool CHyprcursorManager::loadThemeStyle(const SCursorStyleInfo& info) {
                     return false;
                 }
 
-                RsvgRectangle rect = {0, 0, (double)info.size, (double)info.size};
+                RsvgRectangle rect = {0, 0, (double)PIXELSIDE, (double)PIXELSIDE};
 
                 if (!rsvg_handle_render_document(handle, PCAIRO, &rect, &error)) {
                     Debug::log(HC_LOG_ERR, logFn, "Failed rendering svg: {}", error->message);
@@ -558,7 +598,7 @@ void CHyprcursorManager::cursorSurfaceStyleDone(const SCursorStyleInfo& info) {
             const bool isArtificial = e->artificial;
 
             // clean artificial rasters made for this
-            if (isArtificial && e->side == info.size)
+            if (isArtificial && e->side == std::round(info.size / shape->nominalSize))
                 return true;
 
             // clean invalid non-svg rasters
@@ -634,18 +674,22 @@ std::optional<std::string> CHyprcursorImplementation::loadTheme() {
         int         errp = 0;
         zip_t*      zip  = zip_open(cursor.path().string().c_str(), ZIP_RDONLY, &errp);
 
-        zip_file_t* meta_file = zip_fopen(zip, "meta.hl", ZIP_FL_UNCHANGED);
-        bool        metaIsHL  = true;
-        if (!meta_file) {
-            meta_file = zip_fopen(zip, "meta.toml", ZIP_FL_UNCHANGED);
-            metaIsHL  = false;
-            if (!meta_file)
-                return "cursor" + cursor.path().string() + "failed to load meta";
+        zip_int64_t index    = zip_name_locate(zip, "meta.hl", ZIP_FL_ENC_GUESS);
+        bool        metaIsHL = true;
+
+        if (index == -1) {
+            index    = zip_name_locate(zip, "meta.toml", ZIP_FL_ENC_GUESS);
+            metaIsHL = false;
         }
 
-        char* buffer = new char[1024 * 1024]; /* 1MB should be more than enough */
+        if (index == -1)
+            return "cursor" + cursor.path().string() + "failed to load meta";
 
-        int   readBytes = zip_fread(meta_file, buffer, 1024 * 1024 - 1);
+        zip_file_t* meta_file = zip_fopen_index(zip, index, ZIP_FL_UNCHANGED);
+
+        char*       buffer = new char[1024 * 1024]; /* 1MB should be more than enough */
+
+        int         readBytes = zip_fread(meta_file, buffer, 1024 * 1024 - 1);
 
         zip_fclose(meta_file);
 
@@ -670,6 +714,9 @@ std::optional<std::string> CHyprcursorImplementation::loadTheme() {
 
         SHAPE->overrides = meta.parsedData.overrides;
 
+        zip_stat_t sb;
+        zip_stat_init(&sb);
+
         for (auto& i : SHAPE->images) {
             if (SHAPE->shapeType == SHAPE_INVALID) {
                 if (i.filename.ends_with(".svg"))
@@ -694,14 +741,23 @@ std::optional<std::string> CHyprcursorImplementation::loadTheme() {
             IMAGE->delay = i.delay;
             IMAGE->isSVG = SHAPE->shapeType == SHAPE_SVG;
 
-            // read from zip
-            zip_file_t* image_file = zip_fopen(zip, i.filename.c_str(), ZIP_FL_UNCHANGED);
-            if (!image_file)
+            index = zip_name_locate(zip, i.filename.c_str(), ZIP_FL_ENC_GUESS);
+            if (index == -1)
                 return "cursor" + cursor.path().string() + "failed to load image_file";
 
-            IMAGE->data = new char[1024 * 1024]; /* 1MB should be more than enough, again. This probably should be in the spec. */
+            // read from zip
+            zip_file_t* image_file = zip_fopen_index(zip, index, ZIP_FL_UNCHANGED);
+            zip_stat_index(zip, index, ZIP_FL_UNCHANGED, &sb);
 
-            IMAGE->dataLen = zip_fread(image_file, IMAGE->data, 1024 * 1024 - 1);
+            if (sb.valid & ZIP_STAT_SIZE) {
+                IMAGE->data    = new char[sb.size + 1];
+                IMAGE->dataLen = sb.size + 1;
+            } else {
+                IMAGE->data    = new char[1024 * 1024]; /* 1MB should be more than enough, again. This probably should be in the spec. */
+                IMAGE->dataLen = 1024 * 1024;
+            }
+
+            IMAGE->dataLen = zip_fread(image_file, IMAGE->data, IMAGE->dataLen - 1);
 
             zip_fclose(image_file);
 
@@ -712,7 +768,7 @@ std::optional<std::string> CHyprcursorImplementation::loadTheme() {
                 IMAGE->cairoSurface = cairo_image_surface_create_from_png_stream(::readPNG, IMAGE);
 
                 if (const auto STATUS = cairo_surface_status(IMAGE->cairoSurface); STATUS != CAIRO_STATUS_SUCCESS) {
-                    delete[](char*) IMAGE->data;
+                    delete[] (char*)IMAGE->data;
                     IMAGE->data = nullptr;
                     return "Failed reading cairoSurface, status " + std::to_string((int)STATUS);
                 }
@@ -724,10 +780,11 @@ std::optional<std::string> CHyprcursorImplementation::loadTheme() {
         if (SHAPE->images.empty())
             return "meta invalid: no images for shape " + cursor.path().stem().string();
 
-        SHAPE->directory  = cursor.path().stem().string();
-        SHAPE->hotspotX   = meta.parsedData.hotspotX;
-        SHAPE->hotspotY   = meta.parsedData.hotspotY;
-        SHAPE->resizeAlgo = stringToAlgo(meta.parsedData.resizeAlgo);
+        SHAPE->directory   = cursor.path().stem().string();
+        SHAPE->hotspotX    = meta.parsedData.hotspotX;
+        SHAPE->hotspotY    = meta.parsedData.hotspotY;
+        SHAPE->nominalSize = meta.parsedData.nominalSize;
+        SHAPE->resizeAlgo  = stringToAlgo(meta.parsedData.resizeAlgo);
 
         zip_discard(zip);
     }
